@@ -1,11 +1,11 @@
-# $Id: AnyDBD.pm,v 1.9 2000/11/22 12:07:04 matt Exp $
+# $Id: AnyDBD.pm,v 1.12 2001/08/02 16:32:13 matt Exp $
 
 package DBIx::AnyDBD;
 use DBI;
 use strict;
 use vars qw/$AUTOLOAD $VERSION/;
 
-$VERSION = '1.98';
+$VERSION = '2.00';
 
 sub new {
     my $class = shift;
@@ -21,8 +21,8 @@ sub new {
                     RaiseError => 1,
                 })
             );
-    die "Can't connect: ", DBI->errstr unless $dbh;
-    my $package = $args{'package'} || __PACKAGE__;
+    die "Can't connect: " . DBI->errstr unless $dbh;
+    my $package = $args{'package'} || $class;
     my $self = bless { 'package' => $package, dbh => $dbh }, $class;
     $self->rebless;
     $self->_init if $self->can('_init');
@@ -32,10 +32,9 @@ sub new {
 sub connect {
     my $class = shift;
     my ($dsn, $user, $pass, $args, $package) = @_;
-    $dsn ||= $ENV{DBI_DSN} || $ENV{DBI_DBNAME} || '';
     my $dbh = DBI->connect($dsn, $user, $pass, $args);
     return undef unless $dbh;
-    $package ||= __PACKAGE__;
+    $package ||= $class;
     my $self = bless { 'package' => $package, 'dbh' => $dbh }, $class;
     $self->rebless;
     $self->_init if $self->can('_init');
@@ -45,6 +44,10 @@ sub connect {
 sub rebless {
     my $self = shift;
     my $driver = ucfirst($self->{dbh}->{Driver}->{Name});
+    if ( $driver eq 'Proxy' ) {
+	# Looking into the internals of DBD::Proxy is maybe a little questionable
+	($driver) = $self->{dbh}->{proxy_client}->{application} =~ /^DBI:(.+?):/;
+    }
     my $class = $self->{'package'};
     my ($odbc, $ado) = ($driver eq 'ODBC', $driver eq 'ADO');
     if ($odbc || $ado) {
@@ -78,6 +81,8 @@ sub rebless {
 #             }
         elsif ($name eq 'Adaptive Server Anywhere') {
             $driver = 'ASAny';
+        } elsif ($name eq 'ADABAS D') {
+            $driver = 'AdabasD';
         }
         else {  # this should catch Access and Informix
             $driver = lc($name);
@@ -153,24 +158,49 @@ sub load_module {
 }
 
 sub get_dbh {
-    my $self = shift;
     # maybe add code here to check connection status.
     # or maybe add check once every 10 get_dbh's...
-    return $self->{dbh};
+    return shift->{dbh};
 }
 
 sub DESTROY {
     my $self = shift;
-    $self->{dbh}->disconnect;
+    my $dbh;
+    if ($dbh = $self->get_dbh) {
+        $dbh->disconnect;
+    }
 }
 
 sub AUTOLOAD {
-    my $self = shift;
-    my $func = $AUTOLOAD;
-    $func =~ s/.*:://;
+    (my $func = $AUTOLOAD) =~ s/.*:://;
     no strict 'refs';
-    *{$AUTOLOAD} = sub { shift->get_dbh->$func(@_); };
-    return $self->get_dbh->$func(@_);
+    # The following is much more elegant but doesn't work!
+    # *$func = sub {
+    #	  unshift @_ => shift->get_dbh;
+    #	  goto &{$_[0]->can($func)};
+    # };    
+    *$func = sub {
+      my $dbh = shift->get_dbh;
+      if (wantarray) {
+	  my @r = eval { $dbh->$func(@_) };
+	  return @r unless $@;
+      } elsif (defined wantarray) {
+	  my $r = eval { $dbh->$func(@_) };
+	  return $r unless $@;
+      } else {
+	  eval { $dbh->$func(@_) };
+	  return unless $@;
+      };
+      if ( $@ =~ /(.*) at ${\__FILE__} .*$/ ) {
+	  # We want croak() to report errors as occouring
+	  # in the implementation class even though we are related
+	  package DBIx::AnyDBD::Carp;
+	  require Carp;
+	  Carp::croak "$1";
+      }
+      die; # Unreachable ?
+    };
+    goto &$func;
 }
 
 1;
@@ -242,6 +272,11 @@ MyPackage::Oracle::func, and then MyPackage::Default::func. Beware that the
 DBD driver will be ucfirst'ed, because lower case package names are reserved
 as pragmas in perl. See the known DBD package mappings below.
 
+If package parameter is undefined then the package name used to call
+the constructor is used.  This will usually be DBIx::AnyDBD.  This, in
+itself, is not very useful but is convenient if you subclass
+DBIx::AnyDBD.
+
 If attr is undefined then the default attributes are:
 
     AutoCommit => 1
@@ -283,6 +318,7 @@ them:
     DBD::Pg => Pg.pm
     DBD::mysql => Mysql.pm
     DBD::Informix => Informix.pm
+    DBD::AdabasD => AdabasD.pm
 
 If you use this on other platforms, let me know what the mappings are.
 
@@ -298,6 +334,7 @@ The following are the known package mappings for ODBC:
     Microsoft Access => Access.pm
     Informix => Informix.pm
     Oracle => Oracle.pm
+    Adabas D => AdabasD.pm
 
 Anything that isn't listed above will get mapped using the following rule:
 
@@ -329,6 +366,37 @@ connections go through ODBC, but if you're doing some of that funky stuff
 with ADO (such as queries on MS Index Server) then you're not likely to
 need this module!
 
+=head1 Controlling error propagation from AUTOLOADed DBI methods
+
+Typicially the implementation packages will make calls to DBI methods
+as though they were methods of the DBIx::AnyDBD object.  If one of
+these methods reports an error in DBI::AnyDBD then the error is caught
+and rethrown by DBIx::AnyDBD so that the error is reported as occuring
+in the implementation module.  It does this by calling Carp::croak()
+with the current package set to DBIx::AnyDBD::Carp.
+
+Usually this the the right thing to do but sometimes you may want to
+report the error in the line containing the original method call on
+the DBIx::AnyDBD object.  In this case you should temporarily set
+@DBIx::AnyDBD::Carp::ISA.
+
+    my $db = DBIx::AnyDBD->connect("dbi:Oracle:sid1", 
+        "user", "pass", {}, "MyClass");
+
+    my $foo = $db->foo;
+    my $blee = $db->blee("too few arguments"); # Error reported here
+
+    package MyClass::Oracle;
+    
+    sub foo { 
+	shift->prepare("Invalid SQL"); # Error reported here
+    }
+
+    sub blee {
+	local @DBIx::AnyDBD::Carp::ISA = __PACKAGE__;
+	shift->selectall_arrayref(BLEE_STATEMENT,{},@_); # Error not reported here
+    }
+
 =head1 LICENCE
 
 This module is free software, and you may distribute it under the same 
@@ -340,3 +408,9 @@ Commercial support for this module is available on a pay per incident
 basis from Fastnet Software Ltd. Contact matt@sergeant.org for further
 details. Alternatively join the DBI-Users mailing list, where I'll help
 you out for free!
+
+=head1 SEE ALSO
+
+Check out the example files in the example/ directory.
+
+=cut
